#include "sde.h"

// Model variables
Lookup* _rel_attractiveness_lookup;
double _average_delivery_delay;
double _average_life_product;
double _average_orders[2];
double _average_orders_them;
double _average_orders_us;
double _backlog[2];
double _backlog_correction[2];
double _capacity[2];
double _capacity_adjustment[2];
double _capacity_life[2];
double _capacity_them;
double _capacity_us;
double _committal_producer_fraction[2];
double _committals;
double _completions;
double _contact_rate;
double _customer_prevalence;
double _customer_with_non_customer_contacts;
double _customers;
double _delivery_delay[2];
double _desired_backlog[2];
double _desired_capacity[2];
double _desired_production[2];
double _final_time;
double _initial_customers;
double _initial_time;
double _investment[2];
double _new_backlog[2];
double _new_orders[2];
double _new_shipments[2];
double _non_customer_contacts;
double _normal_delivery_delay;
double _orders_received[2];
double _potential_customers;
double _potential_production[2];
double _product_in_use[2];
double _product_per_customer;
double _production[2];
double _reductions[2];
double _relative_attractiveness[2];
double _replacement_backlog[2];
double _replacement_investment[2];
double _replacement_orders[2];
double _replacement_shipments[2];
double _sales_fraction;
double _saveper;
double _target_delivery_delay[2];
double _time_step;
double _time_to_adjust_capacity[2];
double _time_to_adjust_capacity_them;
double _time_to_adjust_capacity_us;
double _time_to_average_orders[2];
double _time_to_correct_backlog[2];
double _total_backlog[2];
double _total_market;
double _total_orders[2];
double _total_shipments[2];
double _waiting_customers;

// Internal variables
const int numOutputs = 5;

// Array dimensions
const size_t _producer[2] = { 0, 1 };

// Dimension mappings


// Internal state
bool lookups_initialized = false;

void initConstants() {
  // Initialize constants.
  // FINAL TIME = 10
  _final_time = 10.0;
  // INITIAL TIME = 0
  _initial_time = 0.0;
  // TIME STEP = 0.0625
  _time_step = 0.0625;
  // average life product = 2
  _average_life_product = 2.0;
  // capacity life[producer] = 2
  for (size_t i = 0; i < 2; i++) {
  _capacity_life[_producer[i]] = 2.0;
  }
  // contact rate = 500
  _contact_rate = 500.0;
  // initial customers = 100000
  _initial_customers = 100000.0;
  // normal delivery delay = 0.25
  _normal_delivery_delay = 0.25;
  // product per customer = 1
  _product_per_customer = 1.0;
  // sales fraction = 0.005
  _sales_fraction = 0.005;
  // target delivery delay[producer] = 0.1
  for (size_t i = 0; i < 2; i++) {
  _target_delivery_delay[_producer[i]] = 0.1;
  }
  // time to adjust capacity them = 1
  _time_to_adjust_capacity_them = 1.0;
  // time to adjust capacity us = 1
  _time_to_adjust_capacity_us = 1.0;
  // time to average orders[producer] = 0.25
  for (size_t i = 0; i < 2; i++) {
  _time_to_average_orders[_producer[i]] = 0.25;
  }
  // time to correct backlog[producer] = 0.5
  for (size_t i = 0; i < 2; i++) {
  _time_to_correct_backlog[_producer[i]] = 0.5;
  }

  // Initialize lookups.
if (!lookups_initialized) {
  _rel_attractiveness_lookup = __new_lookup(4, 0.0, 5.0, 1.0, 1.0, 2.0, 0.5, 5.0, 0.0);

  lookups_initialized = true;
}
}
void initLevels() {
  // Initialize variables with initialization values, such as levels, and the variables they depend on.
  _time = _initial_time;
  // relative attractiveness[producer] = ACTIVE INITIAL(rel attractiveness lookup(delivery delay[producer]/average delivery delay),1)
  for (size_t i = 0; i < 2; i++) {
  _relative_attractiveness[_producer[i]] = 1.0;
  }
  // committal producer fraction[producer] = relative attractiveness[producer]/SUM(relative attractiveness[producer!])
  for (size_t i = 0; i < 2; i++) {
  double __t1 = 0.0;
	for (size_t v = 0; v < 2; v++) {
	  __t1 += _relative_attractiveness[_producer[v]];
	}
  _committal_producer_fraction[_producer[i]] = _relative_attractiveness[_producer[i]] / __t1;
  }
  // Customers = INTEG(completions,initial customers)
  _customers = _initial_customers;
  // Product In Use[producer] = INTEG(new shipments[producer]+replacement shipments[producer]-replacement orders[producer],Customers*committal producer fraction[producer]*product per customer)
  for (size_t i = 0; i < 2; i++) {
  _product_in_use[_producer[i]] = _customers * _committal_producer_fraction[_producer[i]] * _product_per_customer;
  }
  // replacement orders[producer] = Product In Use[producer]/average life product
  for (size_t i = 0; i < 2; i++) {
  _replacement_orders[_producer[i]] = _product_in_use[_producer[i]] / _average_life_product;
  }
  // Replacement Backlog[producer] = INTEG(replacement orders[producer]-replacement shipments[producer],replacement orders[producer]*normal delivery delay)
  for (size_t i = 0; i < 2; i++) {
  _replacement_backlog[_producer[i]] = _replacement_orders[_producer[i]] * _normal_delivery_delay;
  }
  // Potential Customers = INTEG(-committals,1e+07)
  _potential_customers = 10000000.0;
  // total market = ACTIVE INITIAL(Potential Customers+Waiting Customers+Customers,Potential Customers)
  _total_market = _potential_customers;
  // customer prevalence = Customers/total market
  _customer_prevalence = _customers / _total_market;
  // non customer contacts = Potential Customers*contact rate
  _non_customer_contacts = _potential_customers * _contact_rate;
  // customer with non customer contacts = non customer contacts*customer prevalence
  _customer_with_non_customer_contacts = _non_customer_contacts * _customer_prevalence;
  // committals = customer with non customer contacts*sales fraction
  _committals = _customer_with_non_customer_contacts * _sales_fraction;
  // new orders[producer] = committals*product per customer*committal producer fraction[producer]
  for (size_t i = 0; i < 2; i++) {
  _new_orders[_producer[i]] = _committals * _product_per_customer * _committal_producer_fraction[_producer[i]];
  }
  // New Backlog[producer] = INTEG(new orders[producer]-new shipments[producer],new orders[producer]*normal delivery delay)
  for (size_t i = 0; i < 2; i++) {
  _new_backlog[_producer[i]] = _new_orders[_producer[i]] * _normal_delivery_delay;
  }
  // Waiting Customers = INTEG(committals-completions,SUM(New Backlog[producer!])/product per customer)
  double __t2 = 0.0;
	for (size_t v = 0; v < 2; v++) {
	  __t2 += _new_backlog[_producer[v]];
	}
  _waiting_customers = __t2 / _product_per_customer;
  // total orders[producer] = new orders[producer]+replacement orders[producer]
  for (size_t i = 0; i < 2; i++) {
  _total_orders[_producer[i]] = _new_orders[_producer[i]] + _replacement_orders[_producer[i]];
  }
  // orders received[producer] = total orders[producer]
  for (size_t i = 0; i < 2; i++) {
  _orders_received[_producer[i]] = _total_orders[_producer[i]];
  }
  // Average Orders[producer] = INTEG((orders received[producer]-Average Orders[producer])/time to average orders[producer],orders received[producer])
  for (size_t i = 0; i < 2; i++) {
  _average_orders[_producer[i]] = _orders_received[_producer[i]];
  }
  // desired backlog[producer] = Average Orders[producer]*target delivery delay[producer]
  for (size_t i = 0; i < 2; i++) {
  _desired_backlog[_producer[i]] = _average_orders[_producer[i]] * _target_delivery_delay[_producer[i]];
  }
  // Backlog[producer] = INTEG(orders received[producer]-production[producer],desired backlog[producer])
  for (size_t i = 0; i < 2; i++) {
  _backlog[_producer[i]] = _desired_backlog[_producer[i]];
  }
  // backlog correction[producer] = (Backlog[producer]-desired backlog[producer])/time to correct backlog[producer]
  for (size_t i = 0; i < 2; i++) {
  _backlog_correction[_producer[i]] = (_backlog[_producer[i]] - _desired_backlog[_producer[i]]) / _time_to_correct_backlog[_producer[i]];
  }
  // desired production[producer] = Average Orders[producer]+backlog correction[producer]
  for (size_t i = 0; i < 2; i++) {
  _desired_production[_producer[i]] = _average_orders[_producer[i]] + _backlog_correction[_producer[i]];
  }
  // desired capacity[producer] = desired production[producer]
  for (size_t i = 0; i < 2; i++) {
  _desired_capacity[_producer[i]] = _desired_production[_producer[i]];
  }
  // Capacity[producer] = INTEG(investment[producer]-reductions[producer],desired capacity[producer])
  for (size_t i = 0; i < 2; i++) {
  _capacity[_producer[i]] = _desired_capacity[_producer[i]];
  }
}

void evalAux() {
  // Evaluate auxiliaries in order from the bottom up.
  // SAVEPER = TIME STEP
  _saveper = _time_step;
  // average orders them = Average Orders[THEM]
  _average_orders_them = _average_orders[1];
  // average orders us = Average Orders[US]
  _average_orders_us = _average_orders[0];
  // capacity them = Capacity[THEM]
  _capacity_them = _capacity[1];
  // capacity us = Capacity[US]
  _capacity_us = _capacity[0];
  // reductions[producer] = Capacity[producer]/capacity life[producer]
  for (size_t i = 0; i < 2; i++) {
  _reductions[_producer[i]] = _capacity[_producer[i]] / _capacity_life[_producer[i]];
  }
  // potential production[producer] = Capacity[producer]
  for (size_t i = 0; i < 2; i++) {
  _potential_production[_producer[i]] = _capacity[_producer[i]];
  }
  // replacement orders[producer] = Product In Use[producer]/average life product
  for (size_t i = 0; i < 2; i++) {
  _replacement_orders[_producer[i]] = _product_in_use[_producer[i]] / _average_life_product;
  }
  // desired backlog[producer] = Average Orders[producer]*target delivery delay[producer]
  for (size_t i = 0; i < 2; i++) {
  _desired_backlog[_producer[i]] = _average_orders[_producer[i]] * _target_delivery_delay[_producer[i]];
  }
  // backlog correction[producer] = (Backlog[producer]-desired backlog[producer])/time to correct backlog[producer]
  for (size_t i = 0; i < 2; i++) {
  _backlog_correction[_producer[i]] = (_backlog[_producer[i]] - _desired_backlog[_producer[i]]) / _time_to_correct_backlog[_producer[i]];
  }
  // desired production[producer] = Average Orders[producer]+backlog correction[producer]
  for (size_t i = 0; i < 2; i++) {
  _desired_production[_producer[i]] = _average_orders[_producer[i]] + _backlog_correction[_producer[i]];
  }
  // production[producer] = MIN(desired production[producer],potential production[producer])
  for (size_t i = 0; i < 2; i++) {
  _production[_producer[i]] = _MIN(_desired_production[_producer[i]], _potential_production[_producer[i]]);
  }
  // delivery delay[producer] = Backlog[producer]/production[producer]
  for (size_t i = 0; i < 2; i++) {
  _delivery_delay[_producer[i]] = _backlog[_producer[i]] / _production[_producer[i]];
  }
  // average delivery delay = SUM(delivery delay[producer!]*production[producer!])/SUM(production[producer!])
  double __t3 = 0.0;
	for (size_t v = 0; v < 2; v++) {
	  __t3 += _delivery_delay[_producer[v]] * _production[_producer[v]];
	}
  double __t4 = 0.0;
	for (size_t v = 0; v < 2; v++) {
	  __t4 += _production[_producer[v]];
	}
  _average_delivery_delay = __t3 / __t4;
  // relative attractiveness[producer] = ACTIVE INITIAL(rel attractiveness lookup(delivery delay[producer]/average delivery delay),1)
  for (size_t i = 0; i < 2; i++) {
  _relative_attractiveness[_producer[i]] = _LOOKUP(_rel_attractiveness_lookup, _delivery_delay[_producer[i]] / _average_delivery_delay);
  }
  // committal producer fraction[producer] = relative attractiveness[producer]/SUM(relative attractiveness[producer!])
  for (size_t i = 0; i < 2; i++) {
  double __t5 = 0.0;
	for (size_t v = 0; v < 2; v++) {
	  __t5 += _relative_attractiveness[_producer[v]];
	}
  _committal_producer_fraction[_producer[i]] = _relative_attractiveness[_producer[i]] / __t5;
  }
  // total market = ACTIVE INITIAL(Potential Customers+Waiting Customers+Customers,Potential Customers)
  _total_market = _potential_customers + _waiting_customers + _customers;
  // customer prevalence = Customers/total market
  _customer_prevalence = _customers / _total_market;
  // non customer contacts = Potential Customers*contact rate
  _non_customer_contacts = _potential_customers * _contact_rate;
  // customer with non customer contacts = non customer contacts*customer prevalence
  _customer_with_non_customer_contacts = _non_customer_contacts * _customer_prevalence;
  // committals = customer with non customer contacts*sales fraction
  _committals = _customer_with_non_customer_contacts * _sales_fraction;
  // new orders[producer] = committals*product per customer*committal producer fraction[producer]
  for (size_t i = 0; i < 2; i++) {
  _new_orders[_producer[i]] = _committals * _product_per_customer * _committal_producer_fraction[_producer[i]];
  }
  // total orders[producer] = new orders[producer]+replacement orders[producer]
  for (size_t i = 0; i < 2; i++) {
  _total_orders[_producer[i]] = _new_orders[_producer[i]] + _replacement_orders[_producer[i]];
  }
  // orders received[producer] = total orders[producer]
  for (size_t i = 0; i < 2; i++) {
  _orders_received[_producer[i]] = _total_orders[_producer[i]];
  }
  // replacement investment[producer] = reductions[producer]
  for (size_t i = 0; i < 2; i++) {
  _replacement_investment[_producer[i]] = _reductions[_producer[i]];
  }
  // time to adjust capacity[US] = time to adjust capacity us
  _time_to_adjust_capacity[0] = _time_to_adjust_capacity_us;
  // time to adjust capacity[THEM] = time to adjust capacity them
  _time_to_adjust_capacity[1] = _time_to_adjust_capacity_them;
  // desired capacity[producer] = desired production[producer]
  for (size_t i = 0; i < 2; i++) {
  _desired_capacity[_producer[i]] = _desired_production[_producer[i]];
  }
  // capacity adjustment[producer] = (desired capacity[producer]-Capacity[producer])/time to adjust capacity[producer]
  for (size_t i = 0; i < 2; i++) {
  _capacity_adjustment[_producer[i]] = (_desired_capacity[_producer[i]] - _capacity[_producer[i]]) / _time_to_adjust_capacity[_producer[i]];
  }
  // investment[producer] = capacity adjustment[producer]+replacement investment[producer]
  for (size_t i = 0; i < 2; i++) {
  _investment[_producer[i]] = _capacity_adjustment[_producer[i]] + _replacement_investment[_producer[i]];
  }
  // total backlog[producer] = New Backlog[producer]+Replacement Backlog[producer]
  for (size_t i = 0; i < 2; i++) {
  _total_backlog[_producer[i]] = _new_backlog[_producer[i]] + _replacement_backlog[_producer[i]];
  }
  // total shipments[producer] = production[producer]
  for (size_t i = 0; i < 2; i++) {
  _total_shipments[_producer[i]] = _production[_producer[i]];
  }
  // replacement shipments[producer] = total shipments[producer]*Replacement Backlog[producer]/total backlog[producer]
  for (size_t i = 0; i < 2; i++) {
  _replacement_shipments[_producer[i]] = _total_shipments[_producer[i]] * _replacement_backlog[_producer[i]] / _total_backlog[_producer[i]];
  }
  // new shipments[producer] = total shipments[producer]*New Backlog[producer]/total backlog[producer]
  for (size_t i = 0; i < 2; i++) {
  _new_shipments[_producer[i]] = _total_shipments[_producer[i]] * _new_backlog[_producer[i]] / _total_backlog[_producer[i]];
  }
  // completions = SUM(new shipments[producer!])/product per customer
  double __t6 = 0.0;
	for (size_t v = 0; v < 2; v++) {
	  __t6 += _new_shipments[_producer[v]];
	}
  _completions = __t6 / _product_per_customer;
}

void evalLevels() {
  // Evaluate levels.
  // Average Orders[producer] = INTEG((orders received[producer]-Average Orders[producer])/time to average orders[producer],orders received[producer])
  for (size_t i = 0; i < 2; i++) {
  _average_orders[_producer[i]] = _INTEG(_average_orders[_producer[i]], (_orders_received[_producer[i]] - _average_orders[_producer[i]]) / _time_to_average_orders[_producer[i]]);
  }
  // Backlog[producer] = INTEG(orders received[producer]-production[producer],desired backlog[producer])
  for (size_t i = 0; i < 2; i++) {
  _backlog[_producer[i]] = _INTEG(_backlog[_producer[i]], _orders_received[_producer[i]] - _production[_producer[i]]);
  }
  // Capacity[producer] = INTEG(investment[producer]-reductions[producer],desired capacity[producer])
  for (size_t i = 0; i < 2; i++) {
  _capacity[_producer[i]] = _INTEG(_capacity[_producer[i]], _investment[_producer[i]] - _reductions[_producer[i]]);
  }
  // Customers = INTEG(completions,initial customers)
  _customers = _INTEG(_customers, _completions);
  // New Backlog[producer] = INTEG(new orders[producer]-new shipments[producer],new orders[producer]*normal delivery delay)
  for (size_t i = 0; i < 2; i++) {
  _new_backlog[_producer[i]] = _INTEG(_new_backlog[_producer[i]], _new_orders[_producer[i]] - _new_shipments[_producer[i]]);
  }
  // Potential Customers = INTEG(-committals,1e+07)
  _potential_customers = _INTEG(_potential_customers, -_committals);
  // Product In Use[producer] = INTEG(new shipments[producer]+replacement shipments[producer]-replacement orders[producer],Customers*committal producer fraction[producer]*product per customer)
  for (size_t i = 0; i < 2; i++) {
  _product_in_use[_producer[i]] = _INTEG(_product_in_use[_producer[i]], _new_shipments[_producer[i]] + _replacement_shipments[_producer[i]] - _replacement_orders[_producer[i]]);
  }
  // Replacement Backlog[producer] = INTEG(replacement orders[producer]-replacement shipments[producer],replacement orders[producer]*normal delivery delay)
  for (size_t i = 0; i < 2; i++) {
  _replacement_backlog[_producer[i]] = _INTEG(_replacement_backlog[_producer[i]], _replacement_orders[_producer[i]] - _replacement_shipments[_producer[i]]);
  }
  // Waiting Customers = INTEG(committals-completions,SUM(New Backlog[producer!])/product per customer)
  _waiting_customers = _INTEG(_waiting_customers, _committals - _completions);
}

void setInputs(const char* inputData) {
  static double* inputVarPtrs[] = {
    &_time_to_adjust_capacity_them,
    &_time_to_adjust_capacity_us,
  };
  char* inputs = (char*)inputData;
  char* token = strtok(inputs, " ");
  while (token) {
    char* p = strchr(token, ':');
    if (p) {
      *p = '\0';
      int modelVarIndex = atoi(token);
      double value = atof(p+1);
      *inputVarPtrs[modelVarIndex] = value;
    }
    token = strtok(NULL, " ");
  }
}
const char* getHeader() {
  return "Time\taverage orders them\taverage orders us\tcapacity them\tcapacity us";
}

void storeOutputData() {
  outputVar(_time);
  outputVar(_average_orders_them);
  outputVar(_average_orders_us);
  outputVar(_capacity_them);
  outputVar(_capacity_us);
}
